﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace MorphologicalSmoothing
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            string filter = "Image|*.jpg;*.png;*.gif";
            ofd.Filter = filter;
            sfd.Filter = filter;
        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;
        string image_path;
        Bitmap image, processed, display;

        private void processToolStripMenuItem_Click(object sender, EventArgs e)
        {
            processed = processed.MorphologicalSmoothing();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                processed = image;
                display = processed.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static byte[] Erode(this byte[] buffer, BitmapData image_data)
        {
            byte[] result = new byte[buffer.Length];

            for (int x = 1; x < image_data.Width - 1; x++)
            {
                for (int y = 1; y < image_data.Height - 1; y++)
                {
                    int position = x * 3 + y * image_data.Stride;
                    byte val = 255;
                    for (int i = -1; i < 2; i++)
                    {
                        for (int j = -1; j < 2; j++)
                        {
                            int se_pos = position + i * 3 + j * image_data.Stride;
                            val = Math.Min(val, buffer[se_pos]);
                        }
                    }
                    for (int c = 0; c < 3; c++)
                    {
                        result[position + c] = val;
                    }
                }
            }

            return result;
        }

        public static byte[] Dilate(this byte[] buffer, BitmapData image_data)
        {
            byte[] result = new byte[buffer.Length];

            for (int x = 1; x < image_data.Width - 1; x++)
            {
                for (int y = 1; y < image_data.Height - 1; y++)
                {
                    int position = x * 3 + y * image_data.Stride;
                    byte val = 0;
                    for (int i = -1; i < 2; i++)
                    {
                        for (int j = -1; j < 2; j++)
                        {
                            int se_pos = position + i * 3 + j * image_data.Stride;
                            val = Math.Max(val, buffer[se_pos]);
                        }
                    }
                    for (int c = 0; c < 3; c++)
                    {
                        result[position + c] = val;
                    }
                }
            }

            return result;
        }

        public static Bitmap MorphologicalSmoothing(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];
            byte[] result = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            //opening
            result = buffer.Erode(image_data);
            result = result.Dilate(image_data);
            //closing
            result = result.Dilate(image_data);
            result = result.Erode(image_data);

            Bitmap res_img = new Bitmap(w, h);
            BitmapData res_data = res_img.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);
            Marshal.Copy(result, 0, res_data.Scan0, bytes);
            res_img.UnlockBits(res_data);

            return res_img;
        }
    }
}
